@@ -36,7 +36,7 @@
 //! standard. The commands themselves vary based on the application on the
 //! card.
 //!
-//! [6]: http://www.cardwerk.com/smartcards/smartcard_standard_ISO7816-4.aspx
+//! [6]: https://cardwerk.com/iso-7816-part-4/
 //!
 //! ## Note on portability
 //!
@@ -99,10 +99,7 @@
 //!
 //! [8]: https://msdn.microsoft.com/en-us/library/ms953432.aspx#smartcardcspcook_topic2
 #![allow(deprecated)]
-
-#[macro_use]
-extern crate bitflags;
-extern crate pcsc_sys as ffi;
+#![allow(clippy::bad_bit_mask)]
 
 use std::ffi::{CStr, CString};
 use std::mem::{forget, transmute};
@@ -111,6 +108,9 @@ use std::os::raw::c_char;
 use std::ptr::{null, null_mut};
 use std::sync::Arc;
 
+use bitflags::bitflags;
+pub use pcsc_sys as ffi;
+
 use ffi::{DWORD, LONG};
 
 // We use these instead of std::mem::uninitialized -- variables which are
@@ -124,6 +124,8 @@ const DUMMY_DWORD: DWORD = 0xdead_beef;
 
 bitflags! {
     /// A mask of the state a card reader.
+    // Derives to keep backward compat with bitflags 1.
+    #[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy)]
     pub struct State: DWORD {
         const UNAWARE = ffi::SCARD_STATE_UNAWARE;
         const IGNORE = ffi::SCARD_STATE_IGNORE;
@@ -140,6 +142,14 @@ bitflags! {
     }
 }
 
+// Backward compat with bitflags 1.
+impl State {
+    #[deprecated = "use the safe `from_bits_retain` method instead"]
+    pub unsafe fn from_bits_unchecked(bits: DWORD) -> Self {
+        Self::from_bits_retain(bits)
+    }
+}
+
 bitflags! {
     /// A mask of the status of a card in a card reader.
     ///
@@ -148,6 +158,8 @@ bitflags! {
     /// On Windows, Status always has exactly one bit set, and the bit values do
     /// not correspond to underlying PC/SC constants. This allows Status to be
     /// used in the same way across all platforms.
+    // Derives to keep backward compat with bitflags 1.
+    #[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy)]
     pub struct Status: DWORD {
         const UNKNOWN = {
             #[cfg(not(target_os = "windows"))] { ffi::SCARD_UNKNOWN }
@@ -197,6 +209,12 @@ impl Status {
         #[cfg(not(target_os = "windows"))]
         Status::from_bits_truncate(raw_status)
     }
+
+    // Backward compat with bitflags 1.
+    #[deprecated = "use the safe `from_bits_retain` method instead"]
+    pub unsafe fn from_bits_unchecked(bits: DWORD) -> Self {
+        Self::from_bits_retain(bits)
+    }
 }
 
 /// How a reader connection is shared.
@@ -239,6 +257,8 @@ impl Protocol {
 
 bitflags! {
     /// A mask of possible communication protocols.
+    // Derives to keep backward compat with bitflags 1.
+    #[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy)]
     pub struct Protocols: DWORD {
         const UNDEFINED = ffi::SCARD_PROTOCOL_UNDEFINED;
         const T0 = ffi::SCARD_PROTOCOL_T0;
@@ -248,6 +268,14 @@ bitflags! {
     }
 }
 
+// Backward compat with bitflags 1.
+impl Protocols {
+    #[deprecated = "use the safe `from_bits_retain` method instead"]
+    pub unsafe fn from_bits_unchecked(bits: DWORD) -> Self {
+        Self::from_bits_retain(bits)
+    }
+}
+
 /// Disposition method when disconnecting from a card reader.
 #[repr(u32)]
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
@@ -347,6 +375,7 @@ pub enum Error {
 }
 
 impl Error {
+    #[allow(clippy::manual_range_contains)]
     fn from_raw(raw: LONG) -> Error {
         unsafe {
             // The ranges here are the "blocks" above.
@@ -389,7 +418,9 @@ impl std::error::Error for Error {
             Error::NoSmartcard => "The operation requires a Smart Card, but no Smart Card is currently in the device",
             Error::UnknownCard => "The specified smart card name is not recognized",
             Error::CantDispose => "The system could not dispose of the media in the requested manner",
-            Error::ProtoMismatch => "The requested protocols are incompatible with the protocol currently in use with the smart card",
+            Error::ProtoMismatch => {
+                "The requested protocols are incompatible with the protocol currently in use with the smart card"
+            }
             Error::NotReady => "The reader or smart card is not ready to accept commands",
             Error::InvalidValue => "One or more of the supplied parameters values could not be properly interpreted",
             Error::SystemCancelled => "The action was cancelled by the system, presumably to log off or shut down",
@@ -425,14 +456,20 @@ impl std::error::Error for Error {
             Error::CommDataLost => "A communications error with the smart card has been detected. Retry the operation",
             Error::NoKeyContainer => "The requested key container does not exist on the smart card",
             Error::ServerTooBusy => "The smart card resource manager is too busy to complete this operation",
-            Error::UnsupportedCard => "The reader cannot communicate with the card, due to ATR string configuration conflicts",
+            Error::UnsupportedCard => {
+                "The reader cannot communicate with the card, due to ATR string configuration conflicts"
+            }
             Error::UnresponsiveCard => "The smart card is not responding to a reset",
-            Error::UnpoweredCard => "Power has been removed from the smart card, so that further communication is not possible",
+            Error::UnpoweredCard => {
+                "Power has been removed from the smart card, so that further communication is not possible"
+            }
             Error::ResetCard => "The smart card has been reset, so any shared state information is invalid",
             Error::RemovedCard => "The smart card has been removed, so further communication is not possible",
             Error::SecurityViolation => "Access was denied because of a security violation",
             Error::WrongChv => "The card cannot be accessed because the wrong PIN was presented",
-            Error::ChvBlocked => "The card cannot be accessed because the maximum number of PIN entry attempts has been reached",
+            Error::ChvBlocked => {
+                "The card cannot be accessed because the maximum number of PIN entry attempts has been reached"
+            }
             Error::Eof => "The end of the smart card file has been reached",
             Error::CancelledByUser => r#"The user pressed "Cancel" on a Smart Card Selection Dialog"#,
             Error::CardNotAuthenticated => "No PIN was presented to the smart card",
@@ -450,10 +487,12 @@ impl std::fmt::Display for Error {
 }
 
 macro_rules! try_pcsc {
-    ($e:expr) => (match $e {
-        ffi::SCARD_S_SUCCESS => (),
-        err => return Err(Error::from_raw(err)),
-    });
+    ($e:expr) => {
+        match $e {
+            ffi::SCARD_S_SUCCESS => (),
+            err => return Err(Error::from_raw(err)),
+        }
+    };
 }
 
 /// Scope of a context.
@@ -572,7 +611,7 @@ pub fn PNP_NOTIFICATION() -> &'static CStr {
 /// to this function.
 ///
 /// This function wraps the `SCARD_CTL_CODE` macro.
-pub fn ctl_code(code: DWORD) -> DWORD {
+pub const fn ctl_code(code: DWORD) -> DWORD {
     ffi::SCARD_CTL_CODE(code)
 }
 
@@ -671,9 +710,7 @@ impl Context {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gaa1b8970169fd4883a6dc4a8f43f19b67
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379479.aspx
-    pub fn establish(
-        scope: Scope,
-    ) -> Result<Context, Error> {
+    pub fn establish(scope: Scope) -> Result<Context, Error> {
         unsafe {
             let mut handle: ffi::SCARDCONTEXT = DUMMY_LONG as ffi::SCARDCONTEXT;
 
@@ -685,9 +722,7 @@ impl Context {
             ));
 
             Ok(Context {
-                inner: Arc::new(ContextInner {
-                    handle,
-                }),
+                inner: Arc::new(ContextInner { handle }),
             })
         }
     }
@@ -711,15 +746,11 @@ impl Context {
     ///
     /// If the `Context` was cloned, and a clone is still alive, this
     /// function will fail with `Error::CantDispose`.
-    pub fn release(
-        self
-    ) -> Result<(), (Context, Error)> {
+    pub fn release(self) -> Result<(), (Context, Error)> {
         match Arc::try_unwrap(self.inner) {
             Ok(inner) => {
                 unsafe {
-                    let err = ffi::SCardReleaseContext(
-                        inner.handle,
-                    );
+                    let err = ffi::SCardReleaseContext(inner.handle);
                     if err != ffi::SCARD_S_SUCCESS {
                         let context = Context { inner: Arc::new(inner) };
                         return Err((context, Error::from_raw(err)));
@@ -730,7 +761,7 @@ impl Context {
 
                     Ok(())
                 }
-            },
+            }
             Err(arc_inner) => {
                 let context = Context { inner: arc_inner };
                 Err((context, Error::CantDispose))
@@ -745,13 +776,9 @@ impl Context {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga722eb66bcc44d391f700ff9065cc080b
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379788.aspx
-    pub fn is_valid(
-        &self
-    ) -> Result<(), Error> {
+    pub fn is_valid(&self) -> Result<(), Error> {
         unsafe {
-            try_pcsc!(ffi::SCardIsValidContext(
-                self.inner.handle,
-            ));
+            try_pcsc!(ffi::SCardIsValidContext(self.inner.handle,));
 
             Ok(())
         }
@@ -765,13 +792,9 @@ impl Context {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gaacbbc0c6d6c0cbbeb4f4debf6fbeeee6
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379470.aspx
-    pub fn cancel(
-        &self,
-    ) -> Result<(), Error> {
+    pub fn cancel(&self) -> Result<(), Error> {
         unsafe {
-            try_pcsc!(ffi::SCardCancel(
-                self.inner.handle,
-            ));
+            try_pcsc!(ffi::SCardCancel(self.inner.handle,));
 
             Ok(())
         }
@@ -793,29 +816,32 @@ impl Context {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga93b07815789b3cf2629d439ecf20f0d9
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379793.aspx
-    pub fn list_readers<'buf>(
-        &self,
-        buffer: &'buf mut [u8],
-    ) -> Result<ReaderNames<'buf>, Error> {
+    pub fn list_readers<'buf>(&self, buffer: &'buf mut [u8]) -> Result<ReaderNames<'buf>, Error> {
         unsafe {
             assert!(buffer.len() <= std::u32::MAX as usize);
             let mut buflen = buffer.len() as DWORD;
+            // SCardListReaders treats null specially, to query the needed
+            // buffer length. We don't want the caller to be able to trigger
+            // this here -- should use `list_readers_len` instead. So needs
+            // some special treatment.
+            let bufptr = if buflen == 0 {
+                null_mut()
+            } else {
+                let ptr = buffer.as_mut_ptr() as *mut c_char;
+                assert!(!ptr.is_null());
+                ptr
+            };
 
-            let err = ffi::SCardListReaders(
-                self.inner.handle,
-                null(),
-                buffer.as_mut_ptr() as *mut c_char,
-                &mut buflen,
-            );
+            let err = ffi::SCardListReaders(self.inner.handle, null(), bufptr, &mut buflen);
             if err == Error::NoReadersAvailable.into_raw() {
-                return Ok(ReaderNames {
-                    buf: b"\0",
-                    pos: 0,
-                });
+                return Ok(ReaderNames { buf: b"\0", pos: 0 });
             }
             if err != ffi::SCARD_S_SUCCESS {
                 return Err(Error::from_raw(err));
             }
+            if bufptr.is_null() {
+                return Err(Error::InsufficientBuffer);
+            }
 
             Ok(ReaderNames {
                 buf: &buffer[..buflen as usize],
@@ -830,18 +856,11 @@ impl Context {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga93b07815789b3cf2629d439ecf20f0d9
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379793.aspx
-    pub fn list_readers_len(
-        &self,
-    ) -> Result<usize, Error> {
+    pub fn list_readers_len(&self) -> Result<usize, Error> {
         unsafe {
             let mut buflen = DUMMY_DWORD;
 
-            let err = ffi::SCardListReaders(
-                self.inner.handle,
-                null(),
-                null_mut(),
-                &mut buflen,
-            );
+            let err = ffi::SCardListReaders(self.inner.handle, null(), null_mut(), &mut buflen);
             if err == Error::NoReadersAvailable.into_raw() {
                 return Ok(0);
             }
@@ -862,7 +881,11 @@ impl Context {
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga93b07815789b3cf2629d439ecf20f0d9
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379793.aspx
     pub fn list_readers_owned(&self) -> Result<Vec<CString>, Error> {
-        let mut buffer = vec![0u8; self.list_readers_len()?];
+        let len = self.list_readers_len()?;
+        if len == 0 {
+            return Ok(vec![]);
+        }
+        let mut buffer = vec![0u8; len];
         Ok(self.list_readers(&mut buffer)?.map(ToOwned::to_owned).collect())
     }
 
@@ -874,12 +897,7 @@ impl Context {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga4e515829752e0a8dbc4d630696a8d6a5
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379473.aspx
-    pub fn connect(
-        &self,
-        reader: &CStr,
-        share_mode: ShareMode,
-        preferred_protocols: Protocols,
-    ) -> Result<Card, Error> {
+    pub fn connect(&self, reader: &CStr, share_mode: ShareMode, preferred_protocols: Protocols) -> Result<Card, Error> {
         unsafe {
             let mut handle: ffi::SCARDHANDLE = DUMMY_LONG as ffi::SCARDHANDLE;
             let mut raw_active_protocol: DWORD = DUMMY_DWORD;
@@ -921,20 +939,19 @@ impl Context {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga33247d5d1257d59e55647c3bb717db24
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379773.aspx
-    pub fn get_status_change<D>(
-        &self,
-        timeout: D,
-        readers: &mut [ReaderState],
-    ) -> Result<(), Error>
-        where D: Into<Option<std::time::Duration>> {
+    pub fn get_status_change<D>(&self, timeout: D, readers: &mut [ReaderState]) -> Result<(), Error>
+    where
+        D: Into<Option<std::time::Duration>>,
+    {
         let timeout_ms = match timeout.into() {
             Some(duration) => {
-                let timeout_ms_u64 = duration.as_secs()
+                let timeout_ms_u64 = duration
+                    .as_secs()
                     .saturating_mul(1000)
                     .saturating_add(u64::from(duration.subsec_nanos()) / 1_000_000);
                 std::cmp::min(ffi::INFINITE, timeout_ms_u64 as DWORD)
-            },
-            None => ffi::INFINITE
+            }
+            None => ffi::INFINITE,
         };
 
         unsafe {
@@ -957,9 +974,7 @@ impl Drop for ContextInner {
         unsafe {
             // Error is ignored here; to do proper error handling,
             // release() should be called manually.
-            let _err = ffi::SCardReleaseContext(
-                self.handle,
-            );
+            let _err = ffi::SCardReleaseContext(self.handle);
         }
     }
 }
@@ -983,10 +998,7 @@ unsafe impl Sync for Context {}
 impl ReaderState {
     /// Create a ReaderState for a card reader with a given presumed
     /// state.
-    pub fn new<T: Into<CString>>(
-        name: T,
-        current_state: State,
-    ) -> ReaderState {
+    pub fn new<T: Into<CString>>(name: T, current_state: State) -> ReaderState {
         ReaderState {
             inner: ffi::SCARD_READERSTATE {
                 szReader: name.into().into_raw(),
@@ -1012,6 +1024,11 @@ impl ReaderState {
         &self.inner.rgbAtr[0..self.inner.cbAtr as usize]
     }
 
+    /// The last current state that was set.
+    pub fn current_state(&self) -> State {
+        State::from_bits_truncate(self.inner.dwCurrentState)
+    }
+
     /// The last reported state.
     pub fn event_state(&self) -> State {
         State::from_bits_truncate(self.inner.dwEventState)
@@ -1038,10 +1055,13 @@ impl ReaderState {
 impl Drop for ReaderState {
     fn drop(&mut self) {
         // Reclaim the name and drop it immediately.
-        unsafe { CString::from_raw(self.inner.szReader as *mut c_char) };
+        unsafe { drop(CString::from_raw(self.inner.szReader as *mut c_char)) };
     }
 }
 
+unsafe impl Send for ReaderState {}
+unsafe impl Sync for ReaderState {}
+
 /// Status of a card in a card reader.
 #[derive(Clone, Debug)]
 pub struct CardStatus<'names_buf, 'atr_buf> {
@@ -1083,9 +1103,8 @@ impl<'names_buf, 'atr_buf> CardStatus<'names_buf, 'atr_buf> {
     /// This function panics when connected to a reader directly without an
     /// active protocol. Use `protocol2()` instead if you want to avoid this.
     pub fn protocol(&self) -> Protocol {
-        self.protocol.expect(
-            "pcsc::CardStatus::protocol() does not support direct connections; use protocol2() instead"
-        )
+        self.protocol
+            .expect("pcsc::CardStatus::protocol() does not support direct connections; use protocol2() instead")
     }
 
     /// The current ATR string of the card.
@@ -1137,9 +1156,8 @@ impl CardStatusOwned {
     /// This function panics when connected to a reader directly without an
     /// active protocol. Use `protocol2()` instead if you want to avoid this.
     pub fn protocol(&self) -> Protocol {
-        self.protocol.expect(
-            "pcsc::CardStatus::protocol() does not support direct connections; use protocol2() instead"
-        )
+        self.protocol
+            .expect("pcsc::CardStatus::protocol() does not support direct connections; use protocol2() instead")
     }
 
     /// The current ATR string of the card.
@@ -1159,17 +1177,11 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gaddb835dce01a0da1d6ca02d33ee7d861
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379469.aspx
-    pub fn transaction(
-        &mut self,
-    ) -> Result<Transaction, Error> {
+    pub fn transaction(&mut self) -> Result<Transaction, Error> {
         unsafe {
-            try_pcsc!(ffi::SCardBeginTransaction(
-                self.handle,
-            ));
+            try_pcsc!(ffi::SCardBeginTransaction(self.handle,));
 
-            Ok(Transaction {
-                card: self,
-            })
+            Ok(Transaction { card: self })
         }
     }
 
@@ -1191,20 +1203,14 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gaddb835dce01a0da1d6ca02d33ee7d861
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379469.aspx
-    pub fn transaction2(
-        &mut self,
-    ) -> Result<Transaction, (&mut Self, Error)> {
+    pub fn transaction2(&mut self) -> Result<Transaction, (&mut Self, Error)> {
         unsafe {
-            let err = ffi::SCardBeginTransaction(
-                self.handle,
-            );
+            let err = ffi::SCardBeginTransaction(self.handle);
             if err != ffi::SCARD_S_SUCCESS {
                 return Err((self, Error::from_raw(err)));
             }
 
-            return Ok(Transaction {
-                card: self,
-            })
+            Ok(Transaction { card: self })
         }
     }
 
@@ -1252,15 +1258,9 @@ impl Card {
     /// using `Disposition::ResetCard`; you only need to call this
     /// function if you want to handle errors or use a different
     /// disposition method.
-    pub fn disconnect(
-        mut self,
-        disposition: Disposition,
-    ) -> Result<(), (Card, Error)> {
+    pub fn disconnect(mut self, disposition: Disposition) -> Result<(), (Card, Error)> {
         unsafe {
-            let err = ffi::SCardDisconnect(
-                self.handle,
-                disposition.into_raw(),
-            );
+            let err = ffi::SCardDisconnect(self.handle, disposition.into_raw());
             if err != ffi::SCARD_S_SUCCESS {
                 return Err((self, Error::from_raw(err)));
             }
@@ -1288,10 +1288,8 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gae49c3c894ad7ac12a5b896bde70d0382
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379803.aspx
-    #[deprecated(since="2.3.0", note="Use status2() or status2_owned() instead.")]
-    pub fn status(
-        &self,
-    ) -> Result<(Status, Protocol), Error> {
+    #[deprecated(since = "2.3.0", note = "Use status2() or status2_owned() instead.")]
+    pub fn status(&self) -> Result<(Status, Protocol), Error> {
         unsafe {
             let mut raw_status: DWORD = DUMMY_DWORD;
             let mut raw_protocol: DWORD = DUMMY_DWORD;
@@ -1308,9 +1306,8 @@ impl Card {
 
             let status = Status::from_raw(raw_status);
 
-            let protocol = Protocol::from_raw(raw_protocol).expect(
-                "pcsc::Card::status() does not support direct connections; use status2() instead"
-            );
+            let protocol = Protocol::from_raw(raw_protocol)
+                .expect("pcsc::Card::status() does not support direct connections; use status2() instead");
 
             Ok((status, protocol))
         }
@@ -1377,9 +1374,7 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gae49c3c894ad7ac12a5b896bde70d0382
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379803.aspx
-    pub fn status2_len(
-        &self,
-    ) -> Result<(usize, usize), Error> {
+    pub fn status2_len(&self) -> Result<(usize, usize), Error> {
         unsafe {
             let mut names_len: DWORD = DUMMY_DWORD;
             let mut raw_state: DWORD = DUMMY_DWORD;
@@ -1416,7 +1411,12 @@ impl Card {
         let (reader_names, state, protocol, atr_len) = {
             let card_status = self.status2(&mut names_buffer, &mut atr_buffer)?;
             let reader_names = card_status.reader_names.map(ToOwned::to_owned).collect();
-            (reader_names, card_status.state, card_status.protocol, card_status.atr.len())
+            (
+                reader_names,
+                card_status.state,
+                card_status.protocol,
+                card_status.atr.len(),
+            )
         };
 
         atr_buffer.truncate(atr_len);
@@ -1444,14 +1444,21 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gaacfec51917255b7a25b94c5104961602
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379559.aspx
-    pub fn get_attribute<'buf>(
-        &self,
-        attribute: Attribute,
-        buffer: &'buf mut [u8],
-    ) -> Result<&'buf [u8], Error> {
+    pub fn get_attribute<'buf>(&self, attribute: Attribute, buffer: &'buf mut [u8]) -> Result<&'buf [u8], Error> {
         unsafe {
             assert!(buffer.len() <= std::u32::MAX as usize);
             let mut attribute_len = buffer.len() as DWORD;
+            // SCardGetAttribtreats null specially, to query the needed
+            // buffer length. We don't want the caller to be able to trigger
+            // this here -- should use `get_attribute_len` instead. So needs
+            // some special treatment.
+            let bufptr = if attribute_len == 0 {
+                null_mut()
+            } else {
+                let ptr = buffer.as_mut_ptr();
+                assert!(!ptr.is_null());
+                ptr
+            };
 
             try_pcsc!(ffi::SCardGetAttrib(
                 self.handle,
@@ -1459,6 +1466,9 @@ impl Card {
                 buffer.as_mut_ptr(),
                 &mut attribute_len,
             ));
+            if bufptr.is_null() && attribute_len > 0 {
+                return Err(Error::InsufficientBuffer);
+            }
 
             Ok(&buffer[0..attribute_len as usize])
         }
@@ -1470,10 +1480,7 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gaacfec51917255b7a25b94c5104961602
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379559.aspx
-    pub fn get_attribute_len(
-        &self,
-        attribute: Attribute,
-    ) -> Result<usize, Error> {
+    pub fn get_attribute_len(&self, attribute: Attribute) -> Result<usize, Error> {
         unsafe {
             let mut attribute_len = DUMMY_DWORD;
 
@@ -1497,7 +1504,11 @@ impl Card {
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gaacfec51917255b7a25b94c5104961602
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379559.aspx
     pub fn get_attribute_owned(&self, attribute: Attribute) -> Result<Vec<u8>, Error> {
-        let mut buf = vec![0u8; self.get_attribute_len(attribute)?];
+        let len = self.get_attribute_len(attribute)?;
+        if len == 0 {
+            return Ok(vec![]);
+        }
+        let mut buf = vec![0u8; len];
         let n = self.get_attribute(attribute, &mut buf)?.len();
         buf.truncate(n);
         Ok(buf)
@@ -1509,11 +1520,7 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga060f0038a4ddfd5dd2b8fadf3c3a2e4f
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379801.aspx
-    pub fn set_attribute(
-        &self,
-        attribute: Attribute,
-        attribute_data: &[u8],
-    ) -> Result<(), Error> {
+    pub fn set_attribute(&self, attribute: Attribute, attribute_data: &[u8]) -> Result<(), Error> {
         unsafe {
             assert!(attribute_data.len() <= std::u32::MAX as usize);
 
@@ -1543,14 +1550,47 @@ impl Card {
     ///
     /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga9a2d77242a271310269065e64633ab99
     /// [2]: https://msdn.microsoft.com/en-us/library/aa379804.aspx
-    pub fn transmit<'buf>(
+    pub fn transmit<'buf>(&self, send_buffer: &[u8], receive_buffer: &'buf mut [u8]) -> Result<&'buf [u8], Error> {
+        self.transmit2(send_buffer, receive_buffer).map_err(|(err, _)| err)
+    }
+
+    /// Transmit an APDU command to the card.
+    ///
+    /// This functions works like [transmit](#method.transmit) but the error type is
+    /// `(Error, usize)`.
+    ///
+    /// `receive_buffer` is a buffer that should be large enough to hold
+    /// the APDU response.
+    ///
+    /// Returns a slice into `receive_buffer` containing the APDU
+    /// response.
+    ///
+    /// If `receive_buffer` is not large enough to hold the APDU response,
+    /// `Error::InsufficientBuffer` is returned, and the `usize` value is set to the
+    /// required size.
+    ///
+    /// `usize` value of the error has no meaning for other `Error` values than `Error::InsufficientBuffer`.
+    ///
+    /// **Note** that when `Error::InsufficientBuffer` is returned, the provided command has
+    /// been already effectively executed by the card. Do not treat this as a generic way to
+    /// obtain the size of the response as you may end up issuing commands multiple times
+    /// which can lead to unexpected results. Normally, most operations on standard card
+    /// let you know the expected size of the response in advance. Be sure to only use this
+    /// for commands that may be executed multiple times in a row without changing the state
+    /// of the card.
+    ///
+    /// This function wraps `SCardTransmit` ([pcsclite][1], [MSDN][2]).
+    ///
+    /// [1]: https://pcsclite.apdu.fr/api/group__API.html#ga9a2d77242a271310269065e64633ab99
+    /// [2]: https://msdn.microsoft.com/en-us/library/aa379804.aspx
+    pub fn transmit2<'buf>(
         &self,
         send_buffer: &[u8],
         receive_buffer: &'buf mut [u8],
-    ) -> Result<&'buf [u8], Error> {
-        let active_protocol = self.active_protocol.expect(
-            "pcsc::Card::transmit() does not work with direct connections"
-        );
+    ) -> Result<&'buf [u8], (Error, usize)> {
+        let active_protocol = self
+            .active_protocol
+            .expect("pcsc::Card::transmit() does not work with direct connections");
         let send_pci = get_protocol_pci(active_protocol);
         let recv_pci = null_mut();
         assert!(receive_buffer.len() <= std::u32::MAX as usize);
@@ -1559,7 +1599,7 @@ impl Card {
         unsafe {
             assert!(send_buffer.len() <= std::u32::MAX as usize);
 
-            try_pcsc!(ffi::SCardTransmit(
+            let r = ffi::SCardTransmit(
                 self.handle,
                 send_pci,
                 send_buffer.as_ptr(),
@@ -1567,7 +1607,12 @@ impl Card {
                 recv_pci,
                 receive_buffer.as_mut_ptr(),
                 &mut receive_len,
-            ));
+            );
+
+            match r {
+                ffi::SCARD_S_SUCCESS => (),
+                err => return Err((Error::from_raw(err), receive_len as usize)),
+            }
 
             Ok(&receive_buffer[0..receive_len as usize])
         }
@@ -1628,10 +1673,7 @@ impl Drop for Card {
             //
             // Disposition is hard-coded to ResetCard here; to use
             // another method, disconnect() should be called manually.
-            let _err = ffi::SCardDisconnect(
-                self.handle,
-                Disposition::ResetCard.into_raw(),
-            );
+            let _err = ffi::SCardDisconnect(self.handle, Disposition::ResetCard.into_raw());
         }
     }
 }
@@ -1657,15 +1699,9 @@ impl<'tx> Transaction<'tx> {
     /// transaction using `Disposition::LeaveCard`; you only need to call
     /// this function if you want to handle errors or use a different
     /// disposition method.
-    pub fn end(
-        self,
-        disposition: Disposition,
-    ) -> Result<(), (Transaction<'tx>, Error)> {
+    pub fn end(self, disposition: Disposition) -> Result<(), (Transaction<'tx>, Error)> {
         unsafe {
-            let err = ffi::SCardEndTransaction(
-                self.card.handle,
-                disposition.into_raw(),
-            );
+            let err = ffi::SCardEndTransaction(self.card.handle, disposition.into_raw());
             if err != 0 {
                 return Err((self, Error::from_raw(err)));
             }
@@ -1676,6 +1712,21 @@ impl<'tx> Transaction<'tx> {
             Ok(())
         }
     }
+
+    /// Reconnect to the card.
+    ///
+    /// This function wraps `SCardReconnect` ([pcsclite][1], [MSDN][2]).
+    ///
+    /// [1]: https://pcsclite.apdu.fr/api/group__API.html#gad5d4393ca8c470112ad9468c44ed8940
+    /// [2]: https://msdn.microsoft.com/en-us/library/aa379797.aspx
+    pub fn reconnect(
+        &mut self,
+        share_mode: ShareMode,
+        preferred_protocols: Protocols,
+        initialization: Disposition,
+    ) -> Result<(), Error> {
+        self.card.reconnect(share_mode, preferred_protocols, initialization)
+    }
 }
 
 impl<'tx> Drop for Transaction<'tx> {
@@ -1686,10 +1737,7 @@ impl<'tx> Drop for Transaction<'tx> {
             //
             // Disposition is hard-coded to LeaveCard here; to use
             // another method, end() should be called manually.
-            let _err = ffi::SCardEndTransaction(
-                self.card.handle,
-                Disposition::LeaveCard.into_raw(),
-            );
+            let _err = ffi::SCardEndTransaction(self.card.handle, Disposition::LeaveCard.into_raw());
         }
     }
 }
